/usr/bin/touch /var/lib/texmf/postinstall/texlive-collection-context_rm.lsr
/usr/bin/updmap-sys --nohash --nomkmap \
--disable original-context-symbol.map \

/usr/libexec/update_tlpdb remove \
collection-context \
context-account \
context-algorithmic \
context-animation \
context-annotation \
context-bnf \
context-chromato \
context-cmscbf \
context-cmttbf \
context-construction-plan \
context-cyrillicnumbers \
context-degrade \
context-fancybreak \
context-filter \
context-french \
context-fullpage \
context-gantt \
context-gnuplot \
context-handlecsv \
context-layout \
context-letter \
context-lettrine \
context-mathsets \
context-rst \
context-ruby \
context-simplefonts \
context-simpleslides \
context-title \
context-transliterator \
context-typearea \
context-typescripts \
context-vim \
context-visualcounter \
context \
context.x86_64-cygwin \
jmn \

